import logging
import os
from typing import Optional

import cellxgene_census

from cas.dataset_retrieval.dataset_retriever import (
    DatasetRetriever,
    check_file_exists,
    construct_full_download_path,
    create_directory_if_missing,
)


class CxGDownloader(DatasetRetriever):
    def download_data(
        self, file_name: Optional[str] = None, download_dir: Optional[str] = None
    ) -> str:
        """
        Download an AnnData dataset with the specified ID.

        Args:
            file_name: The name of the file to save the downloaded AnnData. If not provided, the dataset
                will be saved with the dataset_id as the file name. The file_name parameter only represents
                the name of the file and does not support absolute or relative paths.
                Use download_dir to specify the directory.
            download_dir: The directory where the AnnData file will be downloaded. If not provided, the
                current working directory will be used. The full path is constructed by combining this directory with `file_name`.

        Returns:
            str: The full path to the downloaded AnnData dataset.
        """
        default_file_name = f"{self.matrix_id}.h5ad"
        full_download_path = construct_full_download_path(
            file_name, download_dir, default_file_name
        )
        create_directory_if_missing(full_download_path)
        file_exists = check_file_exists(full_download_path)

        if not file_exists:
            logging.info(f"Downloading dataset with ID '{self.matrix_id}'...")
            cellxgene_census.download_source_h5ad(
                self.matrix_id, to_path=full_download_path
            )
            logging.info(f"Download complete. File saved at '{full_download_path}'.")
        return full_download_path



import logging
import os
from typing import Optional

import requests
from tqdm import tqdm

from cas.dataset_retrieval.dataset_retriever import (
    DatasetRetriever,
    check_file_exists,
    construct_full_download_path,
    create_directory_if_missing,
)

logging.basicConfig(level=logging.WARNING)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class HTTPDownloader(DatasetRetriever):
    def download_data(
        self, file_name: Optional[str] = None, download_dir: Optional[str] = None
    ) -> str:
        """
        Download an AnnData dataset with the specified ID.

        Args:
            file_name: The name of the file to save the downloaded AnnData. If not provided, the dataset
                will be saved with the dataset_id as the file name. The file_name parameter only represents
                the name of the file and does not support absolute or relative paths.
                Use download_dir to specify the directory.
            download_dir: The directory where the AnnData file will be downloaded. If not provided, the
                current working directory will be used. The full path is constructed by combining this directory with `file_name`.

        Returns:
            str: The full path to the downloaded AnnData dataset.
        """
        url = self.matrix_id
        raw_matrix_id = self.matrix_id.split("/")[-1].split(".")[0]
        default_file_name = f"{raw_matrix_id}.h5ad"
        full_download_path = construct_full_download_path(
            file_name, download_dir, default_file_name
        )
        create_directory_if_missing(full_download_path)
        file_exists = check_file_exists(full_download_path)

        if not file_exists:
            logging.info(f"Downloading dataset with ID '{raw_matrix_id}'...")
            response = requests.get(url, stream=True)
            HTTPDownloader._download_file(response, full_download_path)
            logging.info(f"Download complete. File saved at '{full_download_path}'.")
        return full_download_path

    @staticmethod
    def _download_file(response, full_download_path):
        # Get the total file size from the response headers (if available)
        total_size = int(response.headers.get("content-length", 0))
        # Initialize the tqdm progress bar
        with tqdm(
            total=total_size,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            desc="Downloading",
        ) as progress_bar, open(full_download_path, "wb") as file:
            HTTPDownloader._write_chunks_to_file(response, file, progress_bar)

    @staticmethod
    def _write_chunks_to_file(response, file, progress_bar, chunk_size=8192):
        for chunk in response.iter_content(chunk_size=chunk_size):
            if chunk:  # filter out keep-alive new chunks
                file.write(chunk)
                # Update the progress bar with the size of the chunk
                progress_bar.update(len(chunk))


import logging
import sys
from typing import Optional

from cap_anndata import read_h5ad

from cas.file_utils import read_json_file
from cas.utils.conversion_utils import (
    ANNOTATIONS,
    CELL_IDS,
    CELL_LABEL,
    LABELSET,
    LABELSET_NAME,
    LABELSETS,
    collect_parent_cell_ids,
    copy_and_update_file_path,
    fetch_anndata,
    reformat_json,
)

# Set up logging
logger = logging.getLogger(__name__)


def merge(
    cas_file_path: str,
    anndata_path: Optional[str],
    validate: bool,
    output_file_name: str,
):
    """
    Tests if CAS json and AnnData are compatible and merges CAS into AnnData if possible.

    Args:
        cas_file_path: The path to the CAS json file.
        anndata_path: The path to the AnnData file.
        validate: Boolean to determine if validation checks will be performed before writing to the output AnnData file.
        output_file_name: Output AnnData file name.

    """
    input_json = read_json_file(cas_file_path)

    merge_cas_object(input_json, anndata_path, validate, output_file_name)


def merge_cas_object(
    input_json: dict,
    anndata_file_path: Optional[str],
    validate: bool,
    output_file_path: str,
    download_dir: Optional[str] = None,
):
    """
    Tests if CAS json and AnnData are compatible and merges CAS into AnnData if possible.

    Args:
        input_json: The CAS json object.
        anndata_file_path: The path to the AnnData file.
        validate: Boolean to determine if validation checks will be performed before writing to the output AnnData file.
        output_file_path: Output AnnData file name.
        download_dir: The directory to download AnnData files.

    """
    if not anndata_file_path:
        anndata_file_path = fetch_anndata(input_json, download_dir)
    anndata_file_path = copy_and_update_file_path(anndata_file_path, output_file_path)

    with read_h5ad(file_path=anndata_file_path, edit=True) as cap_adata:
        cap_adata.read_obs()
        obs = cap_adata.obs
        test_compatibility(obs, input_json, validate)

        cap_adata.read_uns()
        cap_adata.uns["cas"] = reformat_json(input_json)

        cap_adata.overwrite()


def test_compatibility(anndata_obs, input_json, validate):
    """
    Tests if CAS and AnnData can be merged.

     Args:
        anndata_obs: The AnnData obs object.
        input_json: The CAS data json object.
        validate: Boolean to determine if validation checks will be performed before writing to the output AnnData file.
    """
    annotations = input_json[ANNOTATIONS]
    obs_index = set(anndata_obs.axes[0].tolist())
    validate_cell_ids(obs_index, annotations, validate)

    labelsets = input_json[LABELSETS]
    matching_obs_keys = get_matching_obs_keys(anndata_obs.columns, labelsets)
    check_labelsets(input_json, anndata_obs, matching_obs_keys, validate)


def check_labelsets(cas_json, input_obs, matching_obs_keys, validate):
    annotations = cas_json[ANNOTATIONS]
    derived_cell_ids = collect_parent_cell_ids(cas_json)

    for ann in annotations:
        if ann[LABELSET] in matching_obs_keys:
            anndata_labelset_cell_ids = (
                input_obs.groupby(ann[LABELSET], observed=False)
                .apply(lambda group: set(group.index), include_groups=False)
                .to_dict()
            )
            for cell_label, cell_list in anndata_labelset_cell_ids.items():
                if cell_list == derived_cell_ids.get(
                    str(ann["cell_set_accession"]), set()
                ):
                    handle_matching_labelset(ann, cell_label, input_obs, validate)
                elif cell_label == ann[CELL_LABEL]:
                    if cell_list == set(ann.get(CELL_IDS, [])):
                        handle_matching_labelset(ann, cell_label, input_obs, validate)
                    else:
                        handle_non_matching_labelset(
                            ann, input_obs, validate, derived_cell_ids
                        )


def get_matching_obs_keys(obs_keys, cas_labelsets):
    cas_labelset_names = {item[LABELSET_NAME] for item in cas_labelsets}
    matching_obs_keys = cas_labelset_names.intersection(obs_keys)
    return list(matching_obs_keys)


def handle_matching_labelset(ann, cell_label, input_obs, validate):
    # Used for label changes
    if cell_label != ann[CELL_LABEL]:
        logger.warning(
            f"{ann[CELL_LABEL]} cell ids from CAS match with the cell ids in {cell_label} from anndata, "
            "but they have different cell labels."
        )
        if validate:
            logger.error("Validation failed. Exiting.")
            sys.exit(1)
        # add new category to labelset column
        input_obs[ann[LABELSET]] = input_obs[ann[LABELSET]].cat.add_categories(
            ann[CELL_LABEL]
        )
        # Overwrite the labelset value with CAS labelset
        input_obs.loc[ann[CELL_IDS], ann[LABELSET]] = input_obs.loc[
            ann[CELL_IDS], ann[LABELSET]
        ].map({cell_label: ann[CELL_LABEL]})


def handle_non_matching_labelset(ann, input_obs, validate, derived_cell_ids):
    # Used for hierarchy changes
    logger.warning(
        f"{ann[CELL_LABEL]} cell ids from CAS do not match with the cell ids from anndata. "
        "Please update your CAS json."
    )
    if validate:
        logger.error("Validation failed. Exiting.")
        sys.exit(1)

    # Flush the labelset from anndata
    # input_anndata.obs.loc[list(cell_list), cell_label] = ""
    # Add labelset from CAS to anndata
    cell_ids = derived_cell_ids.get(str(ann["cell_set_accession"]), set())
    input_obs.loc[list(cell_ids), ann[LABELSET]] = str(ann[CELL_LABEL])


def validate_cell_ids(anndata_cell_ids, annotations, validate):
    # Collect cell ids from annotations
    cas_cell_ids = {cell_id for ann in annotations for cell_id in ann.get(CELL_IDS, [])}

    # Validate cas -> anndata
    if not cas_cell_ids <= anndata_cell_ids:
        logger.warning("Not all members of cell ids from cas exist in anndata.")
        if validate:
            logger.error("Validation failed. Exiting.")
            sys.exit(1)

    # Validate anndata -> cas
    if not anndata_cell_ids <= cas_cell_ids:
        logger.warning("Not all members of cell ids from anndata exist in cas.")
        if validate:
            logger.error("Validation failed. Exiting.")
            sys.exit(1)



from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

from anndata import AnnData

from cas.file_utils import read_anndata_file, read_json_file
from cas.utils.conversion_utils import ANNOTATIONS, CELL_IDS, fetch_anndata


def split_anndata_to_file(
    anndata_file_path: Optional[str],
    cas_json_paths: List[str],
    multiple_outputs: bool,
    compression_method: Optional[Literal["gzip", "lzf"]] = "gzip",
):
    """
    Splits an AnnData file into multiple files based on provided CAS JSON files and writes them to disk.

    Args:
        anndata_file_path: Path to the AnnData file.
        cas_json_paths: List of CAS JSON file paths.
        multiple_outputs: If True, outputs multiple files, one for each CAS JSON file; otherwise, outputs a single file.
        compression_method: Compression method utilized in anndata write function. Default is "gzip".

    """
    if not anndata_file_path:
        anndata_file_path = fetch_anndata(
            cas_json_paths[0]
        )  # Assuming all splits are coming from the same CAS JSON.
    adata = read_anndata_file(anndata_file_path)
    cas_list = {
        Path(cas_json).name: read_json_file(cas_json) for cas_json in cas_json_paths
    }
    cas_paths = list(cas_list.keys())

    result = split_anndata(adata, cas_list, multiple_outputs)
    for idx, anndata_item in enumerate(result):
        anndata_item.write_h5ad(
            Path(
                (
                    f"split_{cas_paths[idx].split('.')[0]}.h5ad"
                    if multiple_outputs
                    else "split_anndata.h5ad"
                )
            ),
            compression=compression_method,
        )


def split_anndata(
    adata: AnnData, cas: Dict[str, Dict[str, Any]], multiple_outputs: bool
) -> List[AnnData]:
    """
    Splits an AnnData object into multiple or single AnnData objects based on the provided CAS data.

    Args:
        adata: AnnData object.
        cas: Dictionary representing the CAS data with its file name as keys.
        multiple_outputs: Determines if the output should be multiple AnnData objects or a single one.

    Returns:
        A list of AnnData objects if multiple_outputs is True, otherwise a single AnnData object.

    Raises:
        ValueError: If any required terms do not exist in the CAS data under 'parent_cell_set_name'.
    """

    def get_cell_ids(cas_data: Dict[str, Dict[str, Any]]) -> List[str]:
        """Extract unique cell IDs from CAS data."""
        return list(
            set(
                [
                    cid
                    for cas_obj in cas_data.values()
                    for annotation in cas_obj[ANNOTATIONS]
                    for cid in annotation.get(CELL_IDS, [])
                ]
            )
        )

    if multiple_outputs:
        splitted_anndata_list = []
        for cas_file_name, cas_object in cas.items():
            cell_ids = get_cell_ids({cas_file_name: cas_object})
            mask = adata.obs.index.isin(cell_ids)
            adata_subset = adata[mask, :].to_memory()
            splitted_anndata_list.append(adata_subset)
        return splitted_anndata_list
    else:
        cell_ids = get_cell_ids(cas)
        mask = adata.obs.index.isin(cell_ids)
        adata_subset = adata[mask, :].to_memory()
        return [adata_subset]



import logging
import os
from abc import ABC, abstractmethod
from typing import Optional


class DatasetRetriever(ABC):
    def __init__(self, matrix_id, **kwargs):
        self.matrix_id = matrix_id
        self.kwargs = kwargs

    @abstractmethod
    def download_data(
        self, file_name: Optional[str] = None, download_dir: Optional[str] = None
    ) -> str:
        """
        Download data based on the matrix_id.

        Args:
            file_name: The name of the file where the dataset will be saved. Defaults to None.
            download_dir: The directory of the file where the dataset will be saved. Defaults to None

        Returns:
            str: The path to the downloaded dataset file.
        """
        pass

    @classmethod
    def create(cls, matrix_id: str, **kwargs):
        """
        Factory method to create a dataset downloader based on the matrix_id format.

        Args:
            matrix_id: Identifier or URL for the dataset.
            **kwargs: Additional arguments for downloader initialization.

        Returns:
            DatasetRetriever: An instance of a subclass of DatasetRetriever based on matrix_id type.

        Raises:
            ValueError: If the matrix_id format is unsupported.
        """
        if matrix_id.startswith("cxg_dataset:"):
            from cas.dataset_retrieval.cxg_downloader import CxGDownloader

            return CxGDownloader(matrix_id.split(":")[-1], **kwargs)
        elif matrix_id.startswith(
            "https://datasets.cellxgene.cziscience.com/"
        ) and matrix_id.endswith(".h5ad"):
            from cas.dataset_retrieval.http_downloader import HTTPDownloader

            return HTTPDownloader(matrix_id, **kwargs)
        elif matrix_id.startswith("s3"):
            from cas.dataset_retrieval.s3_downloader import S3Downloader

            return S3Downloader(matrix_id, **kwargs)
        else:
            raise ValueError(f"Unsupported matrix_id: {matrix_id}")


def check_file_exists(file_name: str) -> bool:
    """
    Check if a file exists at the specified path.

    Args:
        file_name: Path to the file.

    Returns:
        True if the file exists, otherwise False.
    """
    anndata_file_path = os.path.abspath(file_name)

    if os.path.exists(anndata_file_path):
        logging.info(f"File '{anndata_file_path}' already exists. Skipping download.")
        return True

    return False


def create_directory_if_missing(file_name: str):
    """
    Ensure the directory for the specified file exists. If it doesn't, create it.

    Args:
        file_name: Path to the file whose directory needs to be checked or created.
    """
    directory = os.path.dirname(file_name)
    if directory and not os.path.exists(directory):
        os.makedirs(directory, exist_ok=True)


def construct_full_download_path(
    file_name: Optional[str], download_dir: Optional[str], default_file_name: str
) -> str:
    """
    Constructs the full file path for the AnnData file based on the provided file name and download directory.

    Args:
        file_name: The name of the file. If None, `default_file_name` will be used.
        download_dir: The directory where the file should be saved. If None, the current directory will be used.
        default_file_name: The default file name to use if `file_name` is not provided.

    Returns:
        str: The full file path combining the directory and file name.
    """
    anndata_file_path = default_file_name if file_name is None else file_name
    full_download_path = os.path.join(download_dir or "", anndata_file_path)

    return full_download_path



import json
import logging
from collections import defaultdict
from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd
from cap_anndata import read_h5ad

from cas.accession.hash_accession_manager import HashAccessionManager, is_hash_accession
from cas.file_utils import (
    read_json_file,
    update_obs,
    update_uns,
    write_dict_to_json_file,
)
from cas.utils.conversion_utils import (
    ANNOTATIONS,
    AUTHOR_ANNOTATION_FIELDS,
    CELL_IDS,
    CELL_LABEL,
    CELLHASH,
    LABELSET,
    LABELSET_NAME,
    LABELSETS,
    collect_parent_cell_ids,
    copy_and_update_file_path,
    fetch_anndata,
    reformat_json,
    convert_complex_type,
)

# Configure logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


def is_list_of_strings(var):
    """
    Check if a value is a list of strings.

    Parameters:
        var (list or any): The value to be checked.

    Returns:
        bool: True if the value is a list containing only string elements,
              False otherwise.

    """
    return isinstance(var, list) and all(isinstance(item, str) for item in var)


def flatten(
    cas_file_path: str,
    anndata_file_path: Optional[str],
    output_file_path: str,
    fill_na: bool,
):
    """
     Processes and integrates information from a JSON file and an AnnData (Annotated Data) file, creating a new AnnData
     object that incorporates the metadata. The resulting AnnData object is then saved to a new file.

    Args:
        cas_file_path: The path to the CAS json file.
        anndata_file_path: The path to the AnnData file.
        output_file_path: Output AnnData file name.
        fill_na: Boolean flag indicating whether to fill missing values in the 'obs' field with pd.NA. If True, missing
                 values will be replaced with pd.NA; if False, they will remain as empty strings.
    """
    input_json = read_json_file(cas_file_path)

    flatten_cas_object(input_json, anndata_file_path, output_file_path, fill_na)


def flatten_cas_object(
    input_json: dict,
    anndata_file_path: Optional[str],
    output_file_path: str,
    fill_na: bool,
):
    """
     Processes and integrates information from a JSON file and an AnnData (Annotated Data) file, creating a new AnnData
     object that incorporates the metadata. The resulting AnnData object is then saved to a new file.

    Args:
        input_json: CAS json file.
        anndata_file_path: The path to the AnnData file.
        output_file_path: Output AnnData file name.
        fill_na: Boolean flag indicating whether to fill missing values in the 'obs' field with pd.NA. If True, missing
                 values will be replaced with pd.NA; if False, they will remain as empty strings.
    """
    if not anndata_file_path:
        anndata_file_path = fetch_anndata(input_json)
    anndata_file_path = copy_and_update_file_path(anndata_file_path, output_file_path)

    annotations = input_json[ANNOTATIONS]
    parent_cell_ids = collect_parent_cell_ids(input_json)

    with read_h5ad(file_path=anndata_file_path, edit=True) as cap_adata:
        cap_adata.read_obs()
        obs = cap_adata.obs
        obs_index = np.array(cap_adata.obs.axes[0].tolist())

        # obs
        flatten_data = process_annotations(
            annotations, obs_index, parent_cell_ids, fill_na
        )
        update_obs(obs, flatten_data)

        # uns
        uns_json = generate_uns_json(input_json)
        cap_adata.read_uns()
        uns = cap_adata.uns
        update_uns(uns, uns_json)

        cap_adata.overwrite()


def process_annotations(annotations, obs_index, parent_cell_ids, fill_na):
    """
    Processes annotations and generates flattened data for obs dataset.

    Args:
        annotations (list): List of annotations.
        obs_index (np.ndarray): Array representing the index of the obs dataset.
        parent_cell_ids (dict): Dictionary containing parent cell ids.
        fill_na (bool):

    Returns:
        dict: Dictionary containing flattened data.
    """
    accession_manager = HashAccessionManager()
    flatten_data = {}
    for ann in annotations:
        cell_ids = ann.get(CELL_IDS, [])
        if not cell_ids:
            cell_ids = parent_cell_ids.get(ann.get("cell_set_accession", []))

        author_annotations = ann.get(AUTHOR_ANNOTATION_FIELDS, {})
        author_annotations.update(
            {
                CELLHASH: ann.get("cell_set_accession")
                if is_hash_accession(ann.get("cell_set_accession", None))
                else accession_manager.generate_accession_id(cell_ids=cell_ids)
            }
        )
        ann[AUTHOR_ANNOTATION_FIELDS] = author_annotations

        if not cell_ids:
            # only happens if data has multi-inheritance (as in basal ganglia data)
            continue
        # Convert cell_ids to a list if it's not already for np.isin
        if not isinstance(cell_ids, list):
            cell_ids = list(cell_ids)
        mask = np.isin(obs_index, cell_ids)

        for k, v in ann.items():
            if k in [CELL_IDS, LABELSET]:
                continue

            key = f"{ann[LABELSET]}--{k}" if k != CELL_LABEL else ann[LABELSET]
            value = ", ".join(
                sorted([str(value) for value in v] if isinstance(v, list) else [str(v)])
            )

            if key not in flatten_data:
                flatten_data[key] = pd.Series("", index=obs_index)
            flatten_data[key].loc[mask] = value
            if fill_na:
                flatten_data[key].loc[~mask] = pd.NA

    # Convert relevant columns to categorical after the loop
    for key in flatten_data:
        # Get unique values and convert the Series to categorical
        unique_values = pd.unique(flatten_data[key])
        unique_values = unique_values[~pd.isna(unique_values)]
        flatten_data[key] = pd.Series(
            pd.Categorical(flatten_data[key], categories=unique_values), index=obs_index
        )
    return flatten_data


def generate_uns_json(input_json):
    """
    Generates a dictionary representing the uns (unstructured) field in an AnnData object from a given JSON input.

    This function processes information from a JSON input and generates a dictionary that represents the uns (unstructured)
    field in an AnnData object. The resulting dictionary can be used to populate the uns field in the AnnData object.

    Args:
        input_json (dict): A dictionary representing the input CAS JSON data containing annotations.

    Returns:
        dict: A dictionary representing the uns (unstructured) field in an AnnData object, ready to be used as input
              for writing to an AnnData file.

    """
    uns_json = {}
    root_keys = list(input_json.keys())
    root_keys.remove(ANNOTATIONS)

    for key in root_keys:
        value = input_json[key]
        if not value:
            continue

        if is_list_of_strings(value):
            uns_json[key] = ", ".join(sorted(value))
        elif isinstance(value, str):
            uns_json[key] = value
        else:
            metadata_json = {}
            for labelset in value:
                metadata_key = labelset.get(LABELSET_NAME, "")
                metadata_json.update({metadata_key: {}})
                for k, v in labelset.items():
                    if k == LABELSET_NAME:
                        continue
                    metadata_json.get(metadata_key, {}).update({k: v})
            uns_json["cellannotation_metadata"] = metadata_json

    uns_json["cas"] = reformat_json(input_json)

    return uns_json


def unflatten(
    json_file_path: Optional[str],
    anndata_file_path: str,
    output_file_path: str,
    output_json_path: str,
):
    """
     Unflatten an Anndata file and save it. Also creates a CAS json file as output.

    Args:
        json_file_path: The path to the CAS json file.
        anndata_file_path: The path to the AnnData file.
        output_file_path: Output AnnData file name.
        output_json_path: Output CAS JSON file name.
    """
    # TODO review the `cas = None` and `cap_adata.read_uns()` logic!!!
    anndata_file_path = copy_and_update_file_path(anndata_file_path, output_file_path)
    cas = None
    if json_file_path:
        with open(json_file_path, "r") as file:
            cas = json.load(file)

    with read_h5ad(file_path=anndata_file_path, edit=True) as cap_adata:
        if not cas:
            cap_adata.read_uns()
            if "cas" not in cap_adata.uns:
                raise KeyError(
                    "uns section does not have a CAS section. Please check the AnnData file or provide a valid CAS file."
                )
            cas = json.loads(cap_adata.uns["cas"])

        cap_adata.read_obs()
        obs = cap_adata.obs
        new_cas = unflatten_obs(obs, cas)

        cap_adata.uns["cas"] = reformat_json(new_cas)
        # Save your changes to a new or the same AnnData file
        cap_adata.overwrite()

        # Write new cas json to file
        write_dict_to_json_file(output_json_path, new_cas)


def unflatten_obs(obs_df: pd.DataFrame, cas_json: Dict[str, Any]) -> Dict[str, Any]:
    """
    Reverse the flattening process to update the "annotations" section in a CAS object.

    Args:
        obs_df: DataFrame containing the flattened obs columns from an AnnData object.
        cas_json: CAS JSON object.

    Returns:
        Updated CAS JSON with revised annotations.
    """
    # Create a dictionary with a list of columns for each labelset and its dataframe
    labelsets = [labelset[LABELSET_NAME] for labelset in cas_json[LABELSETS]]
    obs_columns_by_labelset = {
        labelset: [col for col in obs_df.columns if labelset in col]
        for labelset in labelsets
    }
    filtered_obs_by_labelset = {
        labelset: obs_df[columns]
        for labelset, columns in obs_columns_by_labelset.items()
    }
    # Find all matching cell sets defined by obs
    cas_dict = create_cell_label_lookup(filtered_obs_by_labelset)
    # Check cell set membership and update cas
    updated_cas = update_cas_json(cas_dict, cas_json)
    # Discard flattened obs
    flattened_columns = [
        col
        for labelset, column_list in obs_columns_by_labelset.items()
        for col in column_list
        if "--" in col
    ]
    for flattened_column in flattened_columns:
        obs_df.remove_column(flattened_column)

    return updated_cas


def create_cell_label_lookup(df_dict: Dict[str, pd.DataFrame]) -> dict:
    """
    Create a lookup dictionary for cell labels with corresponding observations.

    Args:
        df_dict: A dictionary of DataFrames keyed by label sets.

    Returns:
        A nested dictionary where keys are cell labels and values are the observations
        from the obs field in the AnnData object.
    """
    accession_manager = HashAccessionManager()
    # Initialize the dictionary with defaultdict for automatic dictionary creation
    nested_dict = defaultdict(lambda: {CELL_IDS: [], AUTHOR_ANNOTATION_FIELDS: {}})

    # Process each DataFrame in the dictionary
    for df_key, df in df_dict.items():
        grouped = df.groupby(df_key, observed=True)

        for key, group in grouped:
            # Append indices of each group to cell_ids
            cell_id_list = group.index.tolist()
            key_pair = f"{df_key}:{key}"
            nested_dict[key_pair][CELL_IDS].extend(cell_id_list)

            # Process each column in the group
            for col in df.columns:
                if col == df_key:
                    # Store the labelset and cell label for the first column
                    nested_dict[key_pair][LABELSET] = col
                    nested_dict[key_pair][CELL_LABEL] = convert_complex_type(
                        group[col].iloc[0]
                    )
                    # Store cellhash
                    cell_hash = accession_manager.generate_accession_id(
                        cell_ids=cell_id_list, labelset=col, suppress_warnings=True
                    )
                    nested_dict[key_pair][AUTHOR_ANNOTATION_FIELDS][
                        CELLHASH
                    ] = cell_hash
                else:
                    # Split annotation columns and store them in annotations
                    annotation_column = col.split("--")[-1]
                    if annotation_column == AUTHOR_ANNOTATION_FIELDS:
                        annotation_dict = json.loads(
                            group[col].iloc[0].replace("'", '"')
                        )
                        filtered_annotation_dict = {
                            k: v for k, v in annotation_dict.items() if k != CELLHASH
                        }
                        nested_dict[key_pair][annotation_column].update(
                            filtered_annotation_dict
                        )
                    else:
                        nested_dict[key_pair][annotation_column] = convert_complex_type(
                            group[col].iloc[0]
                        )

    return {
        key_: dict(value)
        for key, value in nested_dict.items()
        for key_ in (key, value[AUTHOR_ANNOTATION_FIELDS][CELLHASH])
    }


def update_cas_json(
    cas_dict: Dict[str, Dict[str, Any]], cas_json: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Update the annotations in the CAS JSON using the provided lookup dictionary.

    This function checks the CAS JSON annotations against a lookup dictionary. It updates
    annotations where cell labels or cell hashes match and discards mismatches. It also
    adds new annotations from the lookup dictionary that are not in the CAS JSON.

    Args:
        cas_dict: A lookup dictionary where keys are cell labels or hashes, and
                  values are dictionaries with annotation data.
        cas_json: The CAS JSON object containing existing annotations.

    Returns:
        The updated CAS JSON with synchronized annotations based on the lookup dictionary.
    """
    updated_cas_annotations: List[Dict[str, Any]] = []
    remaining_annotations_labels = list(key for key in cas_dict.keys())
    # remaining_annotations_labels = list(key for key in cas_dict.keys() if not is_hash_accession(key))

    for annotation in cas_json[ANNOTATIONS]:
        labelset_cell_label_pair = f"{annotation[LABELSET]}:{annotation[CELL_LABEL]}"
        cas_cellhash = annotation[AUTHOR_ANNOTATION_FIELDS][CELLHASH]

        if labelset_cell_label_pair in cas_dict.keys():
            obs_annotation = cas_dict[labelset_cell_label_pair]
            obs_cellhash = cas_dict[labelset_cell_label_pair][AUTHOR_ANNOTATION_FIELDS][
                CELLHASH
            ]

            # TODO Make validation optional
            if cas_cellhash == obs_cellhash:
                updated_cas_annotations.append(obs_annotation)
            else:
                logging.warning(
                    f"Cell set annotations for {labelset_cell_label_pair} are discarded because cell hashes do not match."
                )
            index_to_remove = remaining_annotations_labels.index(
                labelset_cell_label_pair
            )
            remaining_annotations_labels.pop(index_to_remove)
            remaining_annotations_labels.pop(index_to_remove)
            # remaining_annotations_labels.remove(labelset_cell_label_pair)
        elif cas_cellhash in cas_dict.keys():
            obs_annotation = cas_dict[cas_cellhash]
            updated_cas_annotations.append(obs_annotation)
            logger.warning(
                f"Cell id hashes, {cas_cellhash}, match but cell labels, {labelset_cell_label_pair}->"
                f"{obs_annotation[CELL_LABEL]}, do not. "
                f"Annotations are updated anyway. There might be a change in Cell Label field!"
            )
            index_to_remove = remaining_annotations_labels.index(cas_cellhash)
            remaining_annotations_labels.pop(index_to_remove)
            remaining_annotations_labels.pop(index_to_remove - 1)

    for label in remaining_annotations_labels:
        updated_cas_annotations.append(cas_dict[label])
        logger.warning(f"New cell set has been added with label {label}.")

    return {
        k: (updated_cas_annotations if k == ANNOTATIONS else v)
        for k, v in cas_json.items()
    }
